(defpackage :aoc/2018/06
  (:use :cl :aoc :alexandria)
  (:export
   #:vector-2d
   #:make-vector-2d
   #:vector-2d-x
   #:vector-2d-y
   #:vector-2d-sub
   #:manhattan-distance
   #:sample-points
   #:top-left
   #:bottom-right))

(in-package :aoc/2018/06)

(defun parse-line (s)
  (map 'list #'parse-integer (cl-ppcre:split ", " s)))

(defvar input-data '())
(setf input-data (map 'list #'parse-line (test-input 2018 6)))

(defstruct vector-2d
  (x 0 :type fixnum)
  (y 0 :type fixnum))

(defun vector-2d-sub (p1 p2)
  (make-vector-2d
   :x (- (vector-2d-x p1) (vector-2d-x p2))
   :y (- (vector-2d-y p1) (vector-2d-y p2))))

(defun manhattan-distance (p1 p2)
  (let ((diff (vector-2d-sub p1 p2)))
    (+
     (abs (vector-2d-x diff))
     (abs (vector-2d-y diff)))))


(defvar sample-data '((1 1)
                      (1 6)
                      (8 3)
                      (3 4)
                      (5 5)
                      (8 9)))

(defun make-vectors (data)
  (map 'list (lambda (p) (make-vector-2d :x (first p) :y (second p))) data))

(defvar sample-points (make-vectors sample-data))
(setf sample-points (make-vectors sample-data))

(defvar input-points)
(setf input-points (make-vectors input-data))

(defun top-left (vectors)
  (let ((x-min (apply #'min (map 'list #'vector-2d-x vectors) ))
        (y-min  (apply #'min (map 'list #'vector-2d-y vectors) )))
    (make-vector-2d :x x-min :y y-min)))

(defun bottom-right (vectors)
  (let ((x-max (apply #'max (map 'list #'vector-2d-x vectors) ))
        (y-max  (apply #'max (map 'list #'vector-2d-y vectors) )))
    (make-vector-2d :x x-max :y y-max)))

(defun add-distance (p1 points)
  (mapcar (lambda (p) (list (manhattan-distance p1 p) p)) points))

(defun closest-distance-points (p1 points)
  (reduce
   (lambda (a b)
     (cond
       ((< (first b) (first a)) b)
       ((= (first b) (first a)) (append a (list b)))
       (t a)))
   (add-distance p1 points)))

(defun calc-areas (points tl br)
  (let ((area-table (make-hash-table)))
    
    (defun update-area-table (x y)
      (let* ((p (make-vector-2d :x x :y y))
	     (closest-points (rest (closest-distance-points p points))))
	(if (= 1 (length closest-points))
	    (incf (gethash (first closest-points) area-table 0)))))
    
    (loop for x from (vector-2d-x tl) to (vector-2d-x br)
	  collect (loop for y from (vector-2d-y tl) to (vector-2d-y br)
			do (update-area-table x y)))
    area-table))


(defun remove-closest (p1 points)
  (let ((closest (rest (closest-distance-points p1 points))))
    (reduce (lambda (acc x) (remove x acc)) closest :initial-value points)))

(defun boundary-points (tl br)
  (append
   (loop for x from (vector-2d-x tl) to (vector-2d-x br)
	 collect (make-vector-2d :x x :y (vector-2d-y tl)))
   (loop for x from (vector-2d-x tl) to (vector-2d-x br)
	 collect (make-vector-2d :x x :y (vector-2d-y br)))
   (loop for y from (1+ (vector-2d-y tl)) to (1- (vector-2d-y br))
	 collect (make-vector-2d :x (vector-2d-x tl) :y y))
   (loop for x from (1+ (vector-2d-x tl)) to (1- (vector-2d-x br))
	 collect (make-vector-2d :x x :y (vector-2d-y br)))))

(defun closest-to-boundary (points tl br)
  (remove-duplicates
   (loop for p in (boundary-points tl br)
	 for closest = (rest (closest-distance-points p points))
	 if (= 1 (length closest))
	   collect (first closest))))


(defun closed-areas (points)
  (let* ((tl (top-left points))
	 (br (bottom-right points))
	 (areas (calc-areas points tl br))
	 (open-points (closest-to-boundary points tl br)))
    (loop for pos being the hash-key in areas
	  and area being the hash-value in areas
	  if (not (member pos open-points))
	    collect area)))

(defun part1 (points)
  (reduce #'max (closed-areas points)))

(defun total-distance (p points)
  (reduce (lambda (acc x) (+ acc (manhattan-distance p x))) points :initial-value 0))

(defun safe-area (points max-distance)

  (let ((tl (top-left points))
	(br (bottom-right points)))
    (loop for x from (vector-2d-x tl) to (vector-2d-x br)
	  sum (loop for y from (vector-2d-y tl) to (vector-2d-y br)
		    for p = (make-vector-2d :x x :y y)
		    count (> max-distance (total-distance p points))))))

(defun part2 (points &optional (max-distance 10000))
  (safe-area points max-distance))
