(defpackage :aoc/2024/08
  (:use :cl :aoc :alexandria :trivia :lla)
  (:export
   #:parse-line
   #:parse-input
   #:sample-text
   #:sample-data
   #:sample-data2
   #:part1
   #:part2
   ))

(in-package :aoc/2024/08)



;; a frequency node represents a location on the map
;; associated to a frequence, like an antenna or antinode
(defstruct frequency-node frequency position)




(defun parse-line (line)
  (loop 
    for x from 0 to (1- (length line))
    for c = (aref line x)    
    if (not (or  (eq c #\.) (eq c #\#)))
      collect (list x c)))

(defstruct antenna-map bounds antennas)

(defun parse-input (lines)
  (make-antenna-map
   :bounds
   (vector (length (first lines)) (length lines))
   :antennas
   (loop for y from 0 to (1- (length lines))
	 for line in lines
	 append (mapcar (lambda (a)
			  (make-frequency-node
			   :position (vector (first a) y)
			   :frequency (second a)))
			(parse-line line)))))

(defparameter input-text (test-input 2024 8))
(defparameter input-data (parse-input input-text))

(defparameter sample-text (aoc:split-lines "............
........0...
.....0......
.......0....
....0.......
......A.....
............
............
........A...
.........A..
............
............"))

(defparameter sample-data 
  (parse-input sample-text))




(defun antinode (ant1 ant2)
  (and
   (eq (frequency-node-frequency ant1) (frequency-node-frequency ant2))
   (make-frequency-node
    :frequency (frequency-node-frequency ant1)
    :position (aops:each
	       (lambda (a b) (- (* 2 b) a))
	       (frequency-node-position ant1)
	       (frequency-node-position ant2)
	       ))))


(defun group-by-frequency (fnodes)
  (let ((groups (make-hash-table :test #'eq)))
    (loop for n in fnodes
	  do (push n (gethash (frequency-node-frequency n) groups)))
    groups))

(defun fnode-in-boundsp (map fnode)
  (loop for x across (frequency-node-position fnode)
	for bound across (antenna-map-bounds map)
	always (< -1 x bound))
  )

(defun antennas-antinodes (antennas)
  (loop for a1 in antennas
	append (loop for a2 in antennas
		     if (not (eq a1 a2))
		       collect (antinode a1 a2))))


(defun count-antinodes (map)
  (let ((antennas-by-f (group-by-frequency (antenna-map-antennas map)))
	(antinode-map (make-hash-table :test #'equalp))
	)
    (loop
      for antennas being the hash-values of antennas-by-f
      do (loop for n in (antennas-antinodes antennas)
	       if (and n (fnode-in-boundsp map n))
		 do (incf (gethash (frequency-node-position n) antinode-map 0))))
    (length (hash-table-keys antinode-map)))
  
  )


(defparameter sample-data2 (parse-input (aoc:split-lines "T....#....
...T......
.T....#...
.........#
..#.......
..........
...#......
..........
....#.....
..........
")))

(defun resonant-nodes (map ant1 ant2)
  (unless
      (or (equalp ant1 ant2)
	  (not (equal
		(frequency-node-frequency ant1)
		(frequency-node-frequency ant2))))
    (loop for i from 1
	  for fn = (make-frequency-node
		    :frequency (frequency-node-frequency ant1)
		    :position (aops:each
			       (lambda (a b) (+ a (* i (- b a))))
			       (frequency-node-position ant1)
			       (frequency-node-position ant2)))
	  while (fnode-in-boundsp map fn)
	  collect fn)))

(defun antennas-resonant-nodes (map antennas)
  (loop for a1 in antennas
	append (loop for a2 in antennas
		     append (resonant-nodes map a1 a2))))

(defun count-resonant-nodes (map)
  (let ((antennas-by-f (group-by-frequency (antenna-map-antennas map)))
	(resonant-node-map (make-hash-table :test #'equalp))
	)
    (loop
      for antennas being the hash-values of antennas-by-f
      do (loop for n in (antennas-resonant-nodes map antennas)
	       if (fnode-in-boundsp map n)	       
		 do (incf (gethash (frequency-node-position n) resonant-node-map 0))))
    (length (hash-table-keys resonant-node-map)))
  
  )

(defun part1 (data)
  (format nil "~A" (count-antinodes data)))

(defun part2 (data)
  (format nil "~A" (count-resonant-nodes data)))

(defun solve-day ()
  (format t "part1: ~A~%" (part1 input-data))
  (format t "part2: ~A~%" (part2 input-data)))
