(defpackage :aoc/2024/09
  (:use :cl :aoc :alexandria :trivia :lla)
  (:export
   #:sample-data
   #:sample-data2
   #:part1
   #:part2
   #:blocks-to-fs
   #:make-freespace
   #:make-inode
   #:fs-compact-step
   #:max-file-id
   ))

(in-package :aoc/2024/09)




(defstruct inode id start length)
(defstruct freespace start length)

(defun blocks-to-fs (blocks)
  (loop
    for block in blocks
    for id from 0  
    for use = (if (eq use :free) :file :free)
    for node = (if (eq use :free)
		   (make-inode :id (floor (/ id 2)) :start pos :length block)
		   (make-freespace :start pos :length block))
    with    pos = 0
    until (emptyp blocks)
    collect node
    do (incf pos block))
  )

(defun parse-input (line)
  (blocks-to-fs (mapcar #'digit-char-p (coerce line 'list))))

(defparameter input-text (first (test-input 2024 9)))
(defparameter input-data (parse-input input-text))

(defparameter sample-text "2333133121414131402")
(defparameter sample-data 
  (parse-input sample-text))

(defun strip-freespace-at-end (fs)
  (if (or (emptyp fs) (inode-p (first (last fs))))
      fs
      (strip-freespace-at-end (subseq fs 0 (1- (length fs))))))

(defun coalesce-last-inodes (fs)
  (let ((last2 (last fs 2)))
    (if (and
	 (= 2 (length last2))
	 (every #'inode-p last2)
	 (reduce #'(lambda (a b) (and a b)) (mapcar #'inode-id last2)))
	(append
	 (subseq fs 0 (- (length fs) 2))
	 (list (make-inode
		:id (inode-id (first last2))
		:start (inode-start (first last2))
		:length (reduce #'+ (mapcar #'inode-length last2)))))
	fs)))

(defun max-file-id (fs)
  (loop
    for n in fs
    if (inode-p n)
      maximize (inode-id n)))

(defun fs-move-file (fs inode inode-pos freespace free-pos)
  "moves a file into a freespace. If the file is too large it will move as much as fits."
  (strip-freespace-at-end
   (append
    (subseq fs 0 free-pos)
    ;; move file into freespace
    (list (make-inode
	   :id (inode-id inode)
	   :start (freespace-start freespace)
	   :length (min (freespace-length freespace) (inode-length inode))))
    ;; update freespace if any remaining
    (if (> (freespace-length freespace) (inode-length inode))
	(list (make-freespace
	       :length (- (freespace-length freespace) (inode-length inode))
	       :start (+ (freespace-start freespace) (inode-length inode)))))
    ;; skip to where file was originally
    (subseq fs (1+ free-pos) inode-pos)
    ;; if any of the file remaining update the length
    (if (<  (freespace-length freespace) (inode-length inode))
	(list (make-inode
	       :id (inode-id inode)
	       :start (inode-start inode)
	       :length (- (inode-length inode) (freespace-length freespace)))))
    ;; copy rest of fs
    (subseq fs (1+ inode-pos)))))

(defun fs-compact-step (fs)
  (let* ((first-free (find-if #'freespace-p fs))
	 (free-pos (position first-free fs))
	 (last-inode (find-if #'inode-p (reverse fs)))
	 (inode-pos (position last-inode fs)))
    (if first-free
	(fs-move-file fs last-inode inode-pos first-free free-pos)
	fs)))

(defun fs-compact (fs)
  (loop while (some #'freespace-p fs)
	do (setf fs (fs-compact-step fs)))
  (coalesce-last-inodes fs))


(defun fs-checksum (fs)
  (loop
    for inode in fs
    if (inode-p inode)
      sum (* (inode-length inode)
	     (inode-id inode)
	     (/ (+ (* 2 (inode-start inode)) (inode-length inode) -1) 2))))


(defun fs-optimize-file (fs id)
  (let* ((inode (find-if
		 (lambda (n)
		   (and
		    (inode-p n)
		    (= (inode-id n) id)))
		 fs))
	 (inode-pos (position inode fs))
	 (freespace (find-if
		     (lambda (n)
		       (and
			(freespace-p n)
			(>= (freespace-length n) (inode-length inode))))
		     fs))
	 (freespace-pos (position freespace fs))
	 )
    (if (and freespace (> inode-pos freespace-pos))
	(fs-move-file fs inode inode-pos freespace freespace-pos)
	fs)))

(defun fs-optimize (fs)
  (let ((ids (reverse (mapcar #'inode-id (remove-if-not #'inode-p fs)))))
    (reduce #'fs-optimize-file ids :initial-value fs)))


(defun part1 (data)
  (let ((cfs (fs-compact data)))
    (format nil "~A" (fs-checksum cfs))))

(defun part2 (data)
  (let ((ofs (fs-optimize data)))
    (format nil "~A" (fs-checksum ofs))))

(defun solve-day ()
  (format t "part1: ~A~%" (part1 input-data))
  (format t "part2: ~A~%" (part2 input-data)))
