(defpackage :aoc/2024/12
  (:use :cl :aoc :alexandria :trivia :lla)
  (:export
   #:sample-data
   #:sample-data2
   #:part1
   #:part2
   #:find-same-neighbors
   #:get-plant
   #:find-region
   #:scan-regions
   #:top-side-p
   #:left-side-p
   #:right-side-p
   #:bottom-side-p
   #:sides
   ))

(in-package :aoc/2024/12)


(defun parse-line (line)
  (coerce line 'list))


(defun parse-input (lines)
  (let ((garden (mapcar #'parse-line lines)))
    (make-array (list (length (first garden)) (length garden))
		:initial-contents garden))
  )

(defparameter input-text (test-input 2024 12))
(defparameter input-data (parse-input input-text))

(defparameter sample-text (aoc:split-lines "RRRRIICCFF
RRRRIICCCF
VVRRRCCFFF
VVRCCCJFFF
VVVVCJJCFE
VVIVCCJJEE
VVIIICJJEE
MIIIIIJJEE
MIIISIJEEE
MMMISSJEEE
"))
(defparameter sample-data (parse-input sample-text))

(defun get-plant (plants y x)
  (if (array-in-bounds-p plants y x)
      (aref plants y x)
      nil))

(defun find-same-neighbors (plants x y)
  (let ((plant (get-plant plants y x)))
    (loop
      for dx in '(-1 0 1 0)
      for dy in '(0 -1 0 1)
      for tx = (+ x dx)
      for ty = (+ y dy)
      if (eq plant (get-plant plants ty tx))
	collect (cons tx ty)
      )))

(defun find-region (plants x y)
  (let ((seen (make-array (array-dimensions plants) :initial-element nil))
	(queue (list (cons x y))))
    (loop
      for p = (first queue)
      for x = (car p)
      for y = (cdr p)
      for neighbors = (find-same-neighbors plants x y)
      if (not (aref seen y x))
	collect (cons x y) 
      do (progn
	   (setf (aref seen y x) t)
	   (setf queue (cdr queue)))
      do (loop for p in neighbors
	       unless (aref seen (cdr p) (car p))	       
		 do (push p queue))
	 
      until (emptyp queue)
      )))

(defun scan-regions (plants)
  (let ((region-id 0)
	(garden-map (make-array (array-dimensions plants) :initial-element -1)))
    (loop
      for y from 0 below (array-dimension plants 1)
      do (loop
	   for x from 0 below (array-dimension plants 0)
	   for r =  (aref garden-map y x) 
	   if (< r 0)
	     do (let ((id (incf region-id)))
		  (loop
		    for region in (find-region plants x y)
		    for x = (car region)
		    for y = (cdr region)
		    do (setf (aref garden-map y x) id)))))
    garden-map))

(defun circumference (garden x y)
  (let ((id (aref garden y x)))
    (- 4
       (if (eq id (get-plant garden (1- y) x)) 1 0)
       (if (eq id (get-plant garden (1+ y) x)) 1 0)
       (if (eq id (get-plant garden y (1- x))) 1 0)
       (if (eq id (get-plant garden y (1+ x))) 1 0)
       )))

(defun calculate-info (garden f)
  (let ((info (make-hash-table)))
    (loop
      for y from 0 below (array-dimension garden 1)
      do (loop
	   for x from 0 below (array-dimension garden 0)
	   do (let* ((id (aref garden y x))
		     (d (gethash id info (cons 0 0))))
		(setf (gethash id info) (cons (1+ (car d)) (+ (apply f (list garden x y)) (cdr d))))))
      )
    info))


(defun calculate-costs (info)
  (loop for k being the hash-keys in info
	for v being the hash-values in info
	sum (* (car v) (cdr v))))



(defparameter sample-data 
  (parse-input sample-text))

(defun part1 (data)
  (format nil "~A" (calculate-costs (calculate-info (scan-regions data) #'circumference))))

(defun top-side-p (garden x y)
  (let* ((id (get-plant garden y x))
	 (lu (eq id (get-plant garden (1- y) (1- x))))
	 (up (eq id (get-plant garden (1- y) x)))
	 (left (eq id (get-plant garden y (1- x)))))
    
    (and (or lu (not left)) (not up))
    ))

(defun left-side-p (garden x y)
  (let* ((id (get-plant garden y x))
	 (lu (eq id (get-plant garden (1- y) (1- x))))
	 (up (eq id (get-plant garden (1- y) x)))
	 (left (eq id (get-plant garden y (1- x)))))
    (and (or lu (not up)) (not left))))

(defun right-side-p (garden x y)
  (let* ((id (get-plant garden y x))
	 (ru (eq id (get-plant garden (1- y) (1+ x))))
	 (up (eq id (get-plant garden (1- y) x)))
	 (right (eq id (get-plant garden y (1+ x)))))
    (and (or ru (not up)) (not right))))

(defun bottom-side-p (garden x y)
  (let* ((id (get-plant garden y x))
	 (lb (eq id (get-plant garden (1+ y) (1- x))))
	 (bottom (eq id (get-plant garden (1+ y) x)))
	 (left (eq id (get-plant garden y (1- x)))))
    (and (or lb (not left)) (not bottom))))

(defun sides (garden x y)
  (+
   (if (top-side-p garden x y) 1 0)
   (if (left-side-p garden x y) 1 0)
   (if (right-side-p garden x y) 1 0)
   (if (bottom-side-p garden x y) 1 0)
   ))

(defun part2 (data)
  (format nil "~A" (calculate-costs (calculate-info (scan-regions data) #'sides)))) 

(defun solve-day ()
  (format t "part1: ~A~%" (part1 input-data))
  (format t "part2: ~A~%" (part2 input-data)))

(defun submit ()
  (let ((p1 (part1 input-data))
        (p2 (part2 input-data)))
    (if p1 (submit-part1 2024 12 p1))
    (if p2 (submit-part2 2024 12 p2))))
