(defpackage :aoc/2024/14
  (:use :cl :aoc :alexandria :lla :ppcre)
  (:export
   #:parse-line
   #:make-robot
   #:quadrant
   #:robot-at-position
   #:robot-at-time
   #:safety-factor
   #:sample-text
   #:sample-data
   #:sample-data2
   #:part1
   #:part2
   ))

(in-package :aoc/2024/14)

(defstruct robot px py vx vy)

(defun robot-position (r)
  (vector (robot-px r) (robot-py r)))

(defun robot-velocity (r)
  (vector (robot-vx r) (robot-vy r)))

(defun robot-at-position (r p)
  (let* ((v (robot-velocity r)))
    (make-robot
     :px (aref p 0)
     :py (aref p 1)
     :vx (aref v 0)
     :vy (aref v 1)
     )))

(defun robot-at-time (r n grid)
  (let* ((p (robot-position r))
	 (v (robot-velocity r))
	 (np (aops:vectorize (p v grid)
	       (mod (+ p (* n v)) grid))))
    (make-robot
     :px (aref np 0)
     :py (aref np 1)
     :vx (aref v 0)
     :vy (aref v 1)
     )))

(defun parse-line (line)
  (let ((numbers nil))
    (do-matches-as-strings
	(s "-?\\d+" line)
      (push (parse-integer s) numbers))
    (make-robot
     :px (nth 3 numbers)
     :py (nth 2 numbers)
     :vx (nth 1 numbers)
     :vy (nth 0 numbers)
     )))


(defun parse-input (lines)
  (mapcar #'parse-line lines))

(defparameter input-text (test-input 2024 14))
(defparameter input-data (parse-input input-text))

(defparameter sample-text (aoc:split-lines "p=0,4 v=3,-3
p=6,3 v=-1,-3
p=10,3 v=-1,2
p=2,0 v=2,-1
p=0,0 v=1,3
p=3,0 v=-2,-2
p=7,6 v=-1,-3
p=3,0 v=-1,-2
p=9,3 v=2,3
p=7,3 v=-1,2
p=2,4 v=2,-3
p=9,5 v=-3,-3
"))

(defparameter sample-data 
  (parse-input sample-text))


(defun quadrant (position gridsize)
  (let ((mx (/ (1- (aref gridsize 0)) 2))
	(my (/ (1- (aref gridsize 1)) 2))
	(x (aref position 0))
	(y (aref position 1)))
    (if (or (= x mx) (= y my))
	nil
	(+
	 (if (< x mx) 0 1)
	 (if (< y my) 0 2)))))

(defun robots-at-time (robots n grid)
  (mapcar
   (lambda (r) (robot-at-time r n grid))
   robots))

(defun map-robots (robots gridsize)

  (let ((map (make-array (reverse (coerce gridsize 'list)) :initial-element 0)))
    (loop
      for r in robots
      do (incf (aref map (robot-py r) (robot-px r))))
    map))

(defun show-robots (robots gridsize)
  (defun render (n)
    (if (= n 0) "." (format nil "~A" n)))
  (let ((map (map-robots robots gridsize)))
    (loop
      for y from 0 below (array-dimension map 0)
      do (loop
	   for x from 0 below (array-dimension map 1)
	   do (format t "~A" (render (aref map y x))))
      do (format t "~%"))))

(defun safety-factor (robots grid)
  
  (let ((qs (make-array '(4) :initial-element 0))
	(rp-100 (lambda (r) (robot-at-time r 100 grid))))
    (loop
      for p in (mapcar rp-100 robots)
      for q = (quadrant (robot-position p) grid)
      if q
	do (incf (aref qs q))
      finally 
	 (return (aops:vectorize-reduce #'* (qs) (identity qs))))))

(defun robots-symmetry-score (robots grid)
  (let ((m (map-robots robots grid)))
    (loop
      for y from 0 below (array-dimension m 0)
      sum (loop
	    for x1 from 0 below (/  (1- (array-dimension m 1)) 2)
	    for x2 from (1- (array-dimension m 1)) downto 0
	    sum (if (and
		     (not (zerop (aref m y x1)))
		     (not (zerop (aref m y x2))))
		    1 0)
	    ))))

(defun sqr (x) (* x x))
(defun robots-variance-score (robots grid)
  (let* ((n (length robots))
	 (mx (/ (loop for r in robots sum (robot-px r)) n))
	 (my (/ (loop for r in robots sum (robot-py r)) n))
	 (vx (/ (loop for r in robots sum (sqr (- (robot-px r) mx))) n))
	 (vy (/ (loop for r in robots sum (sqr (- (robot-py r) my))) n)))
    (list (round mx) (round my) (sqrt vx) (sqrt vy))))

(defun find-t-with-symmetry-higher (max-t min-symmetry robots grid)
  (loop for n from 0 to max-t
	for score = (robots-symmetry-score (robots-at-time robots n grid) grid)
	if (> score min-symmetry)
	  collect n))

(defun first-t-with-variance-lower (max-t max-s robots grid)
  (loop for n from 0 to max-t
	for score = (robots-variance-score (robots-at-time robots n grid) grid)
	for sx = (nth 2 score)
	for sy = (nth 3 score)
	until		 (and (< sx max-s) (< sy max-s))
	finally (return (list n sx sy))))

(defun part1 (data &optional (grid #(101 103)))
  (format nil "~A" (safety-factor data grid)))

(defun part2 (data)
  (let ((stats (first-t-with-variance-lower 50000 20 data #(101 103))))
    (format nil "~A" (first stats))))

(defun solve-day ()
  (format t "part1: ~A~%" (part1 input-data))
  (format t "part2: ~A~%" (part2 input-data)))

(defun submit ()
  (let ((p1 (part1 input-data))
        (p2 (part2 input-data)))
    (if p1 (submit-part1 2024 14 p1))
    (if p2 (submit-part2 2024 14 p2))))
