(defpackage :aoc/2024/16
  (:use :cl :aoc :aoc/maze :alexandria :trivia :lla :queues)
  (:export
   #:parse-input
   #:best-score
   #:best-path
   #:solve-part2
   #:sample-data
   #:sample-data2
   #:part1
   #:part2
   ))

(in-package :aoc/2024/16)

(defun parse-line (line)
  (coerce line 'list))


(defun parse-input (lines)
  (make-array
   (list (length lines) (length (first lines)))
   :initial-contents (mapcar #'parse-line lines)))

(defparameter input-text (test-input 2024 16))
(defparameter input-data (parse-input input-text))

(defparameter sample-text (aoc:split-lines "###############
#.......#....E#
#.#.###.#.###.#
#.....#.#...#.#
#.###.#####.#.#
#.#.#.......#.#
#.#.#####.###.#
#...........#.#
###.#.#####.#.#
#...#.....#.#.#
#.#.#.###.#.#.#
#.....#...#.#.#
#.###.#.#.#.#.#
#S..#.....#...#
###############"))

(defun valid-pos-p (map pos)
  (not (eq #\# (get-pos map pos))))

(defstruct reindeer pos (dir #\>))

(defstruct state path score)

(defun next-moves (map reindeer)
  "return conses of next reindeer and cost to get there"
  (let ((d (reindeer-dir reindeer))
	(p (reindeer-pos reindeer)))
    (remove-if-not
     #'(lambda (r) (valid-pos-p map (reindeer-pos (car r))))
     (list
      (cons (make-reindeer :pos (pos-move p d) :dir d) 1)
      (cons (make-reindeer :pos p :dir (rotate-left d)) 1000)
      (cons (make-reindeer :pos p :dir (rotate-right d)) 1000)))))

(defun heuristic (x finish)
  "estimate the distance for a reindeer from given finish"
  (let* ((a (reindeer-pos finish))
	 (b (reindeer-pos x)))
    (+
     (manhattan-distance a b)
     (if (or
	  (eq (pos-x a) (pos-x b))
	  (eq (pos-y a) (pos-y b)))
	 0
	 1000))))

(defun best-path (map)
  (let* ((start (make-reindeer :pos (find-tile map #\S)))
	 (finish (make-reindeer :pos (find-tile map #\E)))
	 (path-finder (make-a-star
		       #'next-moves
		       #'heuristic
		       :finished-p (lambda (a b)
				     (equalp
				      (reindeer-pos a)
				      (reindeer-pos b)))
		       )))
    (funcall path-finder map start finish))
  )

(defun best-score (map)
  (caar (best-path map)))

(defparameter sample-data 
  (parse-input sample-text))

(defun part1 (data)
  (format nil "~A" (best-score data)))


(defun best-paths (map)
  (let* ((start (find-tile map #\S))
	 (finish (find-tile map #\E))
	 (state-comparison (compare-states-f finish))
	 (best (make-hash-table :test #'equalp))
	 (best-score 100000000)
	 (pos-on-best-path (make-hash-table :test #'equalp))
	 (todo (make-queue :priority-queue :compare state-comparison)))
    (qpush todo (make-state
		 :path (list (make-reindeer :pos start :dir #\>))
		 :score 0))    
    (loop
      for s = (qpop todo)
      until (null s)
      for path = (state-path s)
      for score = (state-score s)
      finally (return (loop for k being the hash-keys in pos-on-best-path collect k))
      if (<= score best-score)
	do (if (equalp finish (reindeer-pos (first (state-path s))))
	       (progn
		 (if (< score best-score)
		     (progn
		       (setf best-score score)
		       (clrhash pos-on-best-path)))
		 (loop
		   for p in path
		   do (incf (gethash (reindeer-pos p) pos-on-best-path 0))))
	       (loop
		 for next in (next-moves map (first path))
		 for next-score = (+ score (or  (cdr next) 0))
		 if (>= (gethash (car next) best 1000000000) next-score)
		   do (progn
			(qpush todo
			       (make-state
				:path (cons (car next) (state-path s))
				:score next-score))
			(setf (gethash (car next) best) next-score))))

      )))

(defun show-paths (map paths)
  (let ((nmap (copy-array map)))
    (loop
      for p in paths
      do (set-pos nmap p #\O))
    (show-map nmap)))

(defun solve-part2 (map)
  (length (best-paths map)))

(defun part2 (data)
  (format nil "~A" (solve-part2 data) ))

(defun solve-day ()
  (format t "part1: ~A~%" (part1 input-data))
  (format t "part2: ~A~%" (part2 input-data)))

(defun submit ()
  (let ((p1 (part1 input-data))
        (p2 (part2 input-data)))
    (if p1 (submit-part1 2024 16 p1))
    (if p2 (submit-part2 2024 16 p2))))
