
(defpackage :aoc/2024/21
  (:use :cl :aoc
	:aoc/maze :alexandria :lla :arrow-macros)
  (:export
   #:sample-data
   #:sample-data2
   #:part1
   #:part2
   #:key-to-pos
   #:press-key
   #:type-manual
   #:complexity
   #:combinations
   #:press-key-combinations
   #:path-invalid-p
   #:easiest-to-type
   #:decode
   #:decode3
   #:fewest-presses
   #:fill-leg-table
   ))

(in-package :aoc/2024/21)

(defun parse-line (line)
  line)

(defun parse-input (lines)
  (mapcar #'parse-line lines))

(defparameter input-text (test-input 2024 21))
(defparameter input-data (parse-input input-text))

(defparameter sample-text (aoc:split-lines "029A
980A
179A
456A
379A"))
(defparameter sample-data 
  (parse-input sample-text))

;; numerical pad:
;; +---+---+---+
;; | 7 | 8 | 9 |
;; +---+---+---+
;; | 4 | 5 | 6 |
;; +---+---+---+
;; | 1 | 2 | 3 |
;; +---+---+---+
;;     | 0 | A |
;;     +---+---+
;; directional pad:
;;     +---+---+
;;     | ^ | A |
;; +---+---+---+
;; | < | v | > |
;; +---+---+---+
;;
;; complexity = #moves/actions * numeric part
;;
;; observations:
;; - directional pad looks like an upside down truncated num pad
;; - no questions about the actual keystrokes, only amount
;; - keystrokes = manhattan distance (moves) + action press
;;
;; approach:
;; - ^ = 0; < = 1; v = 2; > = 3   --> dirpad to numpad
;; - 0-9A --> pos
;; - pos differences
;; - manhattan + 1
;; - sum it.


(defun pos-to-key (pos &optional (numeric t))
  (cond
    ((zerop (pos-y pos)) (if numeric
			     (case (pos-x pos)
			       (2 #\A)
			       (1 #\0)
			       (0 #\?))
			     (case (pos-x pos)
			       (2 #\A)
			       (1 #\^)
			       (0 #\?))))
    ((plusp (pos-y pos)) (if numeric
			     (digit-char
			      (+
			       (* 3 (1- (pos-y pos)))
			       (1+ (pos-x pos))))
			     (case (pos-x pos)
			       (2 #\>)
			       (1 #\v)
			       (0 #\<))))
    (t #\? ))
  )

(defun decode (code &optional (numeric t))
  (let ((parts (ppcre:split "A" code))
	(pos (key-to-pos #\A)))
    (coerce
     (loop
       for sc in parts
       do (loop for c across sc do (setf pos (pos-step pos c numeric)))
       collect (pos-to-key pos numeric))
     'string)))

(defun decode3 (code)
  (decode
   (decode
    (decode code nil)
    nil)
   t))

(defun decoden (code n numeric)
  (if (zerop n)
      (decode code numeric)
      (decode (decoden code (1- n) nil) numeric))
  )

(defun key-to-pos (c)
  (cond
    ((eq c #\0) (make-pos :x 1 :y 0))
    ((eq c #\A) (make-pos :x 2 :y 0))
    ((eq c #\^) (make-pos :x 1 :y 0))
    ((eq c #\<) (make-pos :x 0 :y 1))
    ((eq c #\v) (make-pos :x 1 :y 1))
    ((eq c #\>) (make-pos :x 2 :y 1))
    ((digit-char-p c) (make-pos
		       :x (mod (1- (digit-char-p c)) 3)
		       :y (floor (/ (+ 2 (digit-char-p c)) 3))))
    (t nil)))

(defun press-key (start target &optional (numeric t))
  (let ((diff (pos-subtract
	       (key-to-pos target)
	       (key-to-pos start))))
    (coerce
     (append
      (loop for x from 0 below (pos-x diff) collect #\>)
      (loop for y from 0 below (pos-y diff) collect (if numeric #\^ #\v))
      (loop for y from 0 below (- (pos-y diff)) collect (if numeric #\v #\^))
      (loop for x from 0 below (- (pos-x diff)) collect #\<)
      '(#\A))
     'string)))

(defun pos-step (pos c &optional (numeric t))
  (case c
    ((#\< #\>) (pos-move pos c))
    (#\^ (pos-move pos (if numeric #\v #\^)))
    (#\v (pos-move pos (if numeric #\^ #\v)))))

(defun path-invalid-p (start moves &optional (numeric t))
  (loop
    for dir in moves
    for pos = (pos-step (or pos start) dir numeric)
    thereis (and (zerop (pos-x pos)) (zerop (pos-y pos)))))

(defun press-key-combinations (start target &optional (numeric t))
  (let* ((start-pos (key-to-pos start))
	 (diff (pos-subtract
		(key-to-pos target)
		start-pos))
	 (hp  (loop
		for x from 0 below (abs (pos-x diff))
		collect (if (plusp (pos-x diff)) #\> #\<)))
	 (vp  (loop
		for y from 0 below (abs (pos-y diff))
		collect (if (equal numeric (plusp (pos-y diff))) #\^ #\v)))
	 )
    (mapcar
     #'(lambda (moves) (coerce (append moves '(#\A)) 'string))
     (remove-if
      #'(lambda (moves) (path-invalid-p start-pos moves))
      (list
       (append hp vp)
       (append vp hp))))
    )
  )

(defparameter *tl-cache* (make-hash-table :test #'equal))
(defun reset-cache ()
  (setf *tl-cache* (make-hash-table :test #'equal)))

(defun _type-length (code level numeric)
  (if (zerop level)
      (length code)
      (loop
	for part in (list code)
	sum (let ((keys (coerce code 'list)))
	      (loop
		for start in (cons #\A keys)
		for target in keys
		sum (loop
		      for c in (press-key-combinations start target numeric)
		      minimizing (type-length c (1- level) nil)
		      )))
	)
      )
  )

(defun type-length (code level numeric)
  (multiple-value-bind (v found) (gethash (cons code level) *tl-cache*)
    (if found
	v
	(let ((l (_type-length code level numeric)))
	  (setf (gethash (cons code level) *tl-cache*) l) ))))



(defun print-lengths (data)
  (reset-cache)
  (loop
    for n from 1 to 26
    do (format t "~%n: ~A~%" n)
       (loop
	 for code in data
	 do (format t "~15A~%" (type-length code n t))
	 )))

(defparameter dir-map
  (make-array
   '(2 3) :initial-contents
   #(
     #(#\SPACE #\^ #\A)
     #(  #\<   #\v #\>)
     )))

(defparameter num-map
  (make-array
   '(4 3)
   :initial-contents 
   #(
     #(#\7 #\8 #\9)
     #(#\4 #\5 #\6)
     #(#\1 #\2 #\3)
     #(#\SPACE #\0 #\A))))


(defun dir-pos (c)
  (car (loop
	 for x from 0 below (array-dimension dir-map 1)
	 append (loop
		  for y from 0 below (array-dimension dir-map 0)
		  if (eq c (aref dir-map y x))
		    collect (cons x y)))))

(defun num-pos (c)
  (car (loop
	 for x from 0 below (array-dimension num-map 1)
	 append (loop
		  for y from 0 below (array-dimension num-map 0)
		  if (eq c (aref num-map y x))
		    collect (cons x y)))))

(defvar leg-lengths (make-hash-table :test #'equal))

(defun fewest-presses (layer ks)
  (loop
    for ki in (cons #\A (coerce ks 'list))
    for kf in (coerce ks 'list)
    sum (let ((ll (gethash (list layer ki kf) leg-lengths)))
	  ll)))

(defun press-move (d pc mc)
  (loop
    for i from 1 to (abs d)
    collect (if (plusp d) pc mc)))

(defun fill-leg-table (n-robots)
  (setf leg-lengths (make-hash-table :test #'equal))
  
  (loop
    for ki from 0 below (array-total-size dir-map)
    do (loop
	 for kf from 0 below (array-total-size dir-map)
	 do (setf
	     (gethash
	      (list
	       0
	       (row-major-aref dir-map ki)
	       (row-major-aref dir-map kf))
	      leg-lengths)
	     1)
	 ))
  (loop
    for layer from 1 to n-robots
    do (let* ((key-map (if (= layer n-robots) num-map dir-map))
	      (key-pos (if (= layer n-robots) #'num-pos #'dir-pos))
	      (space-pos (funcall key-pos #\SPACE)))
	 (loop
	   for i from 0 below (array-total-size key-map)
	   for ki = (row-major-aref key-map i) 
	   for pos-i = (funcall key-pos ki)
	   for xi = (car pos-i)
	   for yi = (cdr pos-i)
	   do (loop
		for j from 0 below (array-total-size key-map)
		for kf = (row-major-aref key-map j)
		for pos-f = (funcall key-pos kf)
		for xf = (car pos-f)
		for yf = (cdr pos-f)
		do (let* (
			  (dx (- xf xi))
			  (dy (- yf yi))
			  (hp (press-move dx #\> #\<))
			  (vp (press-move dy #\v #\^))
			  (hf (coerce (append hp vp '(#\A)) 'string))
			  (h-first (if (equal (cons xf yi) space-pos)
				       99999999999
				       (fewest-presses (1- layer) hf)))
			  (vf (coerce (append vp hp '(#\A)) 'string))
			  (v-first (if (equal (cons xi yf) space-pos)
				       99999999999
				       (fewest-presses (1- layer) vf))))
		     ;;(format t "h ~A ~A ~A ~A~%" xi yf hf h-first)
		     ;;(format t "v ~A ~A ~A ~A~%" xf yi vf v-first)
		     ;;(format t "ll[~A ~A ~A] ~A~%" layer ki kf (min h-first v-first))
		     (setf
		      (gethash (list layer ki kf) leg-lengths)
		      (min h-first v-first)))))
	 
	 ))
  )


(defun print-leg-lengths (stream)
  (loop
    for k being the hash-keys in leg-lengths
    for v being the hash-values in leg-lengths
    do (format stream "~2,'0D,~A,~A,~A~%" (first k) (second k) (third k) v)))

(defun complexity (code n)
  (let ((l (fewest-presses n code))
	(v (parse-integer (ppcre:scan-to-strings "\\d+" code))))
    (* l v)))


(defun print-llengths (data)
  (reset-cache)
  (loop
    for n from 1 to 26
    do (format t "~%n: ~A~%" n)
       (fill-leg-table n)
       (loop
	 for code in data
	 do (format t "~15A~%" (fewest-presses n code))
	 )))

(defun write-leg-lengths-to-file (filename)
  (with-open-file (str filename
		       :direction :output
		       :if-exists :supersede
		       :if-does-not-exist :create)
    (print-leg-lengths str)))




(defun part1 (data)
  (fill-leg-table 3)
  (loop
    for code in data
    sum (complexity code 3)))

(defun part2 (data)
  (fill-leg-table 26)
  (loop
    for code in data
    sum (complexity code 26)))

(defun solve-day ()
  (reset-cache)
  (format t "part1: ~A~%" (part1 input-data))
  (format t "part2: ~A~%" (part2 input-data)))

(defun submit ()
  (let ((p1 (part1 input-data))
        (p2 (part2 input-data)))
    (if p1 (submit-part1 2024 21 p1))
    (if p2 (submit-part2 2024 21 p2))))
