(defpackage :aoc/maze
  (:use :cl :alexandria :trivia :lla :queues)
  (:export
   #:parse-maze
   #:make-a-star
   #:pos
   #:make-pos
   #:pos-x
   #:pos-y
   #:pos-move
   #:pos-add
   #:pos-subtract
   #:show-map
   #:get-pos
   #:set-pos
   #:find-tile
   #:rotate-left
   #:rotate-right
   #:manhattan-distance
   ))

(in-package :aoc/maze)

(defun parse-line (line)
  (coerce line 'list))


(defun parse-maze (lines)
  (make-array
   (list (length lines) (length (first lines)))
   :initial-contents (mapcar #'parse-line lines)))

(defstruct pos
  (x 0 :type fixnum)
  (y 0 :type fixnum))

(defun pos-move (pos dir)
  (case dir
    (#\^ (make-pos :x (pos-x pos) :y (1- (pos-y pos))))
    (#\v (make-pos :x (pos-x pos) :y (1+ (pos-y pos))))
    (#\< (make-pos :x (1- (pos-x pos)) :y (pos-y pos)))
    (#\> (make-pos :x (1+ (pos-x pos)) :y (pos-y pos)))
    ))

(defun pos-add (a b)
  (make-pos
   :x (+ (pos-x a) (pos-x b))
   :y (+ (pos-y a) (pos-y b))
   ))

(defun pos-subtract (a b)
  (make-pos
   :x (- (pos-x a) (pos-x b))
   :y (- (pos-y a) (pos-y b))
   ))


(defun show-map (map &optional (dest t))
  (let ((w (array-dimension map 1))
	(h (array-dimension map 0)))
    (format
     dest "~{~A~}"
     (loop
       for row from 0 below h
       collect (format nil "~A~%"
		       (coerce
			(loop for col from 0 below w
			      collect (aref map row col))
			'string))))))

(defun get-pos (map pos)
  (aref map (pos-y pos) (pos-x pos)))

(defun set-pos (map pos tile)
  (setf (aref map (pos-y pos) (pos-x pos)) tile))

(defun find-tile (map char)
  (let ((w (array-dimension map 1))
	(h (array-dimension map 0)))
    (loop
      for row from 0 below h
      for col = (loop
		  for col from 0 below w
		  if (eq (aref map row col) char) return col)
      if col
	return (make-pos :x col :y row)
      )))

(defun swap-pos (map p np)
  "swaps content of map. returns np"
  (let ((o (get-pos map p)))
    (setf (aref map (pos-y p) (pos-x p)) (get-pos map np))
    (setf (aref map (pos-y np) (pos-x np)) o)
    np))


(defun manhattan-distance (a b)
  "estimate the distance"
  (+
   (abs (- (pos-x a) (pos-x b)))
   (abs (- (pos-y a) (pos-y b)))
   ))

(defun rotate-left (dir)
  (case dir
    (#\^ #\<)
    (#\v #\>)
    (#\< #\v)
    (#\> #\^)
    ))
(defun rotate-right (dir)
  (case dir
    (#\^ #\>)
    (#\v #\<)
    (#\< #\^)
    (#\> #\v)
    ))

(defun state-score (s)
  (caar s))

(defun state-x (s)
  (cdar s))

(defun make-a-star (next-moves heuristic &key (finished-p #'equalp))
  (lambda (map start finish)
    (defun state-less-p (s1 s2)
      (< (+ (state-score s1)
	    (funcall heuristic (state-x s1) finish))
	 (+ (state-score s2)
	    (funcall heuristic (state-x s2) finish))))
    (let* ((best (make-hash-table :test #'equalp))
	   (todo (queues:make-queue :priority-queue :compare #'state-less-p)))
      (queues:qpush todo (list (cons 0 start)))    
      (loop
	for s = (queues:qpop todo)
	until (or
	       (null s)
	       (funcall finished-p (state-x s) finish))
	finally (return s)
	do (loop
	     for next in (funcall next-moves map (state-x s))
	     for score = (+ (state-score s) (cdr next))
	     for next-x = (car next)	     
	     if (> (gethash next-x best 1000000000) score)
	       do (progn
		    (queues:qpush todo (cons (cons score next-x) s))
		    (setf (gethash next-x best) score)))

	))))

(defun make-a-star-all-best-paths (next-moves heuristic &key (finished-p #'equalp))
  (lambda (map start finish)
    (defun state-less-p (s1 s2)
      (< (+ (state-score s1)
	    (funcall heuristic (state-x s1) finish))
	 (+ (state-score s2)
	    (funcall heuristic (state-x s2) finish))))
    (let* ((best (make-hash-table :test #'equalp))
	   (todo (queues:make-queue :priority-queue :compare #'state-less-p)))
      (queues:qpush todo (list (cons 0 start)))    
      (loop
	for s = (queues:qpop todo)
	until (or
	       (null s)
	       (funcall finished-p (state-x s) finish))
	finally (return s)
	do (loop
	     for next in (funcall next-moves map (state-x s))
	     for score = (+ (state-score s) (cdr next))
	     for next-x = (car next)	     
	     if (>= (gethash next-x best 1000000000) score)
	       do (progn
		    (queues:qpush todo (cons (cons score next-x) s))
		    (setf (gethash next-x best) score)))

	))))
